
//
// The tube structure coordinates
//
#declare g_num_stations = 9;
#declare g_tube_station_coordinates = array[g_num_stations][2];

#declare g_tube_station_coordinates[0][0] = -9;
#declare g_tube_station_coordinates[0][1] = 0.2;

#declare g_tube_station_coordinates[1][0] = -18;
#declare g_tube_station_coordinates[1][1] = -0.8;

#declare g_tube_station_coordinates[2][0] = -28;
#declare g_tube_station_coordinates[2][1] = 0.7;

#declare g_tube_station_coordinates[3][0] = -36;
#declare g_tube_station_coordinates[3][1] = -0.4;

#declare g_tube_station_coordinates[4][0] = -46;
#declare g_tube_station_coordinates[4][1] = -1.3;

#declare g_tube_station_coordinates[5][0] = -63;
#declare g_tube_station_coordinates[5][1] = -0.1;

#declare g_tube_station_coordinates[6][0] = -70;
#declare g_tube_station_coordinates[6][1] = 1.0;

#declare g_tube_station_coordinates[7][0] = -80;
#declare g_tube_station_coordinates[7][1] = -0.2;

#declare g_tube_station_coordinates[8][0] = -93;
#declare g_tube_station_coordinates[8][1] = -0.3;

#macro macro_Unit_Gantry_Beam(p_beam_width, p_beam_breadth, p_translate)
    #local l_beam_side_thickness = 0.1;
    #local l_r = l_beam_width * l_beam_side_thickness / 2;
    #local l_f = l_r;
    union {
        cylinder {
            <-1, -p_beam_width / 2 + l_f, -p_beam_breadth / 2 + l_f>,
            <1, -p_beam_width / 2 + l_f, -p_beam_breadth / 2 + l_f>,
            l_r
        }
        cylinder {
            <-1, p_beam_width / 2 - l_f, -p_beam_breadth / 2 + l_f>,
            <1, p_beam_width / 2 - l_f, -p_beam_breadth / 2 + l_f>,
            l_r
        }
        cylinder {
            <-1, -p_beam_width / 2 + l_f, p_beam_breadth / 2 - l_f>,
            <1, -p_beam_width / 2 + l_f, p_beam_breadth / 2 - l_f>,
            l_r
        }
        cylinder {
            <-1, p_beam_width / 2 - l_f, p_beam_breadth / 2 - l_f>,
            <1, p_beam_width / 2 - l_f, p_beam_breadth / 2 - l_f>,
            l_r
        }
        box {
            <-1, -p_beam_width / 2, -p_beam_breadth / 2 + l_f>,
            <1, -p_beam_width / 2 + 2 * l_f, p_beam_breadth / 2 - l_f>
        }
        box {
            <-1, p_beam_width / 2, -p_beam_breadth / 2 + l_f>,
            <1, p_beam_width / 2 - 2 * l_f, p_beam_breadth / 2 - l_f>
        }
        box {
            <-1, -p_beam_width / 2, l_f>,
            <1, p_beam_width / 2, -l_f>
        }
        translate p_translate
    }
#end

#macro macro_Unit_Gantry_Strut(p_beam_width, p_beam_breadth, p_translate)
    #local l_beam_side_thickness = 0.1;
    #local l_r = l_beam_width * l_beam_side_thickness / 2;
    #local l_f = l_r;
    #local l_to = <p_translate.x, 0, p_translate.z>;
    difference {
        union {
            cylinder {
                p_translate + <-p_beam_width / 2 + l_f, 0, -p_beam_breadth / 2 + l_f>,
                l_to + <-p_beam_width / 2 + l_f, 0, -p_beam_breadth / 2 + l_f>,
                l_r
            }
            cylinder {
                p_translate + <p_beam_width / 2 - l_f, 0, -p_beam_breadth / 2 + l_f>,
                l_to + <p_beam_width / 2 - l_f, 0, -p_beam_breadth / 2 + l_f>,
                l_r
            }
            cylinder {
                p_translate + <-p_beam_width / 2 + l_f, 0, p_beam_breadth / 2 - l_f>,
                l_to + <-p_beam_width / 2 + l_f, 0, p_beam_breadth / 2 - l_f>,
                l_r
            }
            cylinder {
                p_translate + <p_beam_width / 2 - l_f, 0, p_beam_breadth / 2 - l_f>,
                l_to + <p_beam_width / 2 - l_f, 0, p_beam_breadth / 2 - l_f>,
                l_r
            }
            box {
                p_translate + <-p_beam_width / 2, 0, -p_beam_breadth / 2 + l_f>,
                l_to + <-p_beam_width / 2 + 2 * l_f, 0, p_beam_breadth / 2 - l_f>
            }
            box {
                p_translate + <p_beam_width / 2, 0, -p_beam_breadth / 2 + l_f>,
                l_to + <p_beam_width / 2 - 2 * l_f, 0, p_beam_breadth / 2 - l_f>
            }
            box {
                p_translate + <-p_beam_width / 2, 0, l_f>,
                l_to + <p_beam_width / 2, 0, -l_f>
            }
        }
        // The difference vs the cylinder is needed to get the nicely flush
        // top ends of the struts.  As we're working from a unit box here,
        // it's easy to figure out the right size cylinder to match....
        cylinder {
            <-1, 0, 0>
            <1, 0, 0>
            1
        }
    }
#end

#macro macro_Unit_Gantry_Cable(p_cable_width, p_start_position, p_end_position,
                    p_mat_cable)
    cylinder {
        p_start_position,
        p_end_position,
        p_cable_width
        texture {
            p_mat_cable
        }
    }
#end

#macro macro_Unit_Gantry_Platform(p_platform_width, p_platform_breadth,
                p_translate)
    #local l_plate_width = p_platform_width * 0.8;
    #local l_plate_height = p_platform_width * 0.005;
    #local l_trim_radius = p_platform_width * 0.005;
    union {
        box {
            <-p_platform_breadth/2, 0, -l_plate_width/2>
            <p_platform_breadth/2, l_plate_height, l_plate_width/2>
        }
        cylinder {
            <-p_platform_breadth/2, l_plate_height / 2, -l_plate_width/2>
            <-p_platform_breadth/2, l_plate_height / 2, l_plate_width/2>
            l_trim_radius
        }
        cylinder {
            <p_platform_breadth/2, l_plate_height / 2, -l_plate_width/2>
            <p_platform_breadth/2, l_plate_height / 2, l_plate_width/2>
            l_trim_radius
        }
        cylinder {
            <-p_platform_breadth/2, l_plate_height / 2, l_plate_width/2>
            <p_platform_breadth/2, l_plate_height / 2, l_plate_width/2>
            l_trim_radius
        }
        cylinder {
            <-p_platform_breadth/2, l_plate_height / 2, -l_plate_width/2>
            <p_platform_breadth/2, l_plate_height / 2, -l_plate_width/2>
            l_trim_radius
        }
        sphere {
            <p_platform_breadth/2, l_plate_height / 2, l_plate_width/2>
            l_trim_radius
        }
        sphere {
            <p_platform_breadth/2, l_plate_height / 2, -l_plate_width/2>
            l_trim_radius
        }
        sphere {
            <-p_platform_breadth/2, l_plate_height / 2, l_plate_width/2>
            l_trim_radius
        }
        sphere {
            <-p_platform_breadth/2, l_plate_height / 2, -l_plate_width/2>
            l_trim_radius
        }
        translate p_translate
    }
#end

#declare g_cable_backarray = array[2][4];
#declare g_cable_backarray[0][0] = <0, 0, 0>;
#declare g_cable_backarray[0][1] = <0, 0, 0>;
#declare g_cable_backarray[0][2] = <0, 0, 0>;
#declare g_cable_backarray[0][3] = <0, 0, 0>;
#declare g_cable_backarray[1][0] = <0, 0, 0>;
#declare g_cable_backarray[1][1] = <0, 0, 0>;
#declare g_cable_backarray[1][2] = <0, 0, 0>;
#declare g_cable_backarray[1][3] = <0, 0, 0>;

#macro macro_Unit_Gantry(p_begin, p_end, p_radius, p_mat_beam, p_mat_cable)
    // Build a gantry in a unit square then scale, rotate, and translate it
    // to the correct position
    #local l_gap_proportion = 0.4;
    #local l_wire_width = 0.01;
    #local l_beam_width = 0.10;
    #local l_beam_breadth = 0.18;
    #local l_strut_width = 0.05;
    #local l_strut_breadth = 0.18;
    #local l_cable_sag = 0.03;
    #local l_cable_width = vlength(p_begin-p_end) * 0.00075;
    #local l_cable_distance = 0.03;
    #local l_num_struts = 3;
    #local l_num_platforms = 4;

    union {
        union {
            // Horizontal beams
            macro_Unit_Gantry_Beam(l_beam_width, l_beam_breadth,
                <0, -1 + l_beam_width/2, 1 - l_beam_breadth>)
            macro_Unit_Gantry_Beam(l_beam_width, l_beam_breadth,
                <0, -1 + l_beam_width/2, -1 + l_beam_breadth>)
            macro_Unit_Gantry_Beam(l_beam_width, l_beam_breadth,
                <0, -1 * (1 + l_gap_proportion) + l_beam_width/2, 1 - l_beam_breadth>)
            macro_Unit_Gantry_Beam(l_beam_width, l_beam_breadth,
                <0, -1 * (1 + l_gap_proportion) + l_beam_width/2, -1 + l_beam_breadth>)

            // More horizontal beams
            macro_Unit_Gantry_Beam(l_beam_width, l_beam_breadth,
                <0, -1 * (1 + l_gap_proportion) + l_beam_width/2, 1 - l_beam_breadth * 2.7>)
            macro_Unit_Gantry_Beam(l_beam_width, l_beam_breadth,
                <0, -1 * (1 + l_gap_proportion) + l_beam_width/2, -1 + l_beam_breadth * 2.7>)

            // Vertical Struts
            #local l_i = 0;
            #while (l_i < l_num_struts)
                #local l_current_xoffset = -1 + (2 / l_num_struts) * (l_i + 0.5);
                #local l_current_yoffset = -1 * (1 + l_gap_proportion);
                #local l_current_zoffset = 1 * (1 - l_strut_breadth);
    
                macro_Unit_Gantry_Strut(l_strut_width, l_strut_breadth,
                    <l_current_xoffset, l_current_yoffset, l_current_zoffset>)
                macro_Unit_Gantry_Strut(l_strut_width, l_strut_breadth,
                    <l_current_xoffset, l_current_yoffset, -l_current_zoffset>)
    
                #local l_i = l_i + 1;
            #end
    
            texture {
                p_mat_beam
            }
        }

        union {
            // Tension Cables
            #local l_i = 0;
            #local l_current_yoffset = -1 * (1 + l_gap_proportion);
            #local l_current_zoffset = 1 * (1 - l_strut_breadth);
            #while (l_i < l_num_struts)
                #local l_current_xoffset = -1 + (2 / l_num_struts) * (l_i + 0.5);
    
                #local l_next_xoffset = -1 + (2 / l_num_struts) * (l_i + 1.5);
    
                #if (l_next_xoffset < 1)
                    macro_Unit_Gantry_Cable(l_wire_width,
                        <l_current_xoffset, l_current_yoffset, l_current_zoffset>,
                        <l_next_xoffset, -1, l_current_zoffset>,
                        p_mat_cable)
                    macro_Unit_Gantry_Cable(l_wire_width,
                        <l_current_xoffset, -1, l_current_zoffset>,
                        <l_next_xoffset, l_current_yoffset, l_current_zoffset>,
                        p_mat_cable)
                    macro_Unit_Gantry_Cable(l_wire_width,
                        <l_current_xoffset, l_current_yoffset, -l_current_zoffset>,
                        <l_next_xoffset, -1, -l_current_zoffset>,
                        p_mat_cable)
                    macro_Unit_Gantry_Cable(l_wire_width,
                        <l_current_xoffset, -1, -l_current_zoffset>,
                        <l_next_xoffset, l_current_yoffset, -l_current_zoffset>,
                        p_mat_cable)
                #end
    
                #local l_i = l_i + 1;
            #end
            #local l_xinc = (2 / l_num_struts) * 0.5;
            macro_Unit_Gantry_Cable(l_wire_width,
                <-1, l_current_yoffset, l_current_zoffset>,
                <-1 + l_xinc, -1, l_current_zoffset>,
                p_mat_cable)
            macro_Unit_Gantry_Cable(l_wire_width,
                <-1, l_current_yoffset, -l_current_zoffset>,
                <-1 + l_xinc, -1, -l_current_zoffset>,
                p_mat_cable)
            macro_Unit_Gantry_Cable(l_wire_width,
                <1, l_current_yoffset, l_current_zoffset>,
                <1 - l_xinc, -1, l_current_zoffset>,
                p_mat_cable)
            macro_Unit_Gantry_Cable(l_wire_width,
                <1, l_current_yoffset, -l_current_zoffset>,
                <1 - l_xinc, -1, -l_current_zoffset>,
                p_mat_cable)
        }
        macro_Axis_Remap(p_begin, p_end, p_radius)
    }

    // Add hanging cables.
    #if (g_lod_cables > 0)
        #local l_i = 0;
        #local l_current_yoffset = -1 * (1 + l_gap_proportion);
        #local l_current_zoffset = 1 * (1 - l_strut_breadth * 0.75);

        #while (l_i < 4)
            union {
                #local l_point1 = <-2/3, l_current_yoffset, l_current_zoffset>;
                #local l_point2 = <0, l_current_yoffset, l_current_zoffset>;
                #local l_point3 = <2/3, l_current_yoffset, l_current_zoffset>;
                #local l_left = macro_Axis_Remap_Point(l_point1, p_begin, p_end, p_radius);
                #local l_mid = macro_Axis_Remap_Point(l_point2, p_begin, p_end, p_radius);
                #local l_right = macro_Axis_Remap_Point(l_point3, p_begin, p_end, p_radius);
                macro_build_cable3(l_left, l_mid, l_right, l_cable_sag, l_cable_width)
                #if (vlength(l_left - g_cable_backarray[0][l_i]) < vlength(p_begin - p_end))
                    macro_build_cable2(g_cable_backarray[0][l_i], l_left, l_cable_sag, l_cable_width)
                #end
                #declare g_cable_backarray[0][l_i] = l_right;
    
                #local l_point1 = l_point1 * <1, 1, -1>;
                #local l_point2 = l_point2 * <1, 1, -1>;
                #local l_point3 = l_point3 * <1, 1, -1>;
                #local l_left = macro_Axis_Remap_Point(l_point1, p_begin, p_end, p_radius);
                #local l_mid = macro_Axis_Remap_Point(l_point2, p_begin, p_end, p_radius);
                #local l_right = macro_Axis_Remap_Point(l_point3, p_begin, p_end, p_radius);
                macro_build_cable3(l_left, l_mid, l_right, l_cable_sag, l_cable_width)
                #if (vlength(l_left - g_cable_backarray[1][l_i]) < vlength(p_begin - p_end))
                    macro_build_cable2(g_cable_backarray[1][l_i], l_left, l_cable_sag, l_cable_width)
                #end
                #declare g_cable_backarray[1][l_i] = l_right;
                texture {
                    t_Cables[l_i]
                }
            }

            #local l_current_zoffset = l_current_zoffset - l_cable_distance;
            #local l_i = l_i + 1;
        #end
    #end

#end

#macro macro_Sectioned_Spline_Lattice(p_spline, p_minor_radius, p_thickness, p_segments, p_join_pct, p_material1, p_material2)

    #local l_increment = 1 / p_segments;
    #local l_i = 0;
    
    
    union {
        #while (l_i < 1)

            macro_Unit_Gantry(p_spline(l_i), p_spline(l_i + l_increment),
                p_minor_radius, p_material2, p_material2)

            #local l_cyl_begin = p_spline(l_i);
            #local l_cyl_end = p_spline(l_i + l_increment);
            #local l_cyl2_begin = l_cyl_begin + (0.001 * (l_cyl_begin - l_cyl_end));
            #local l_cyl2_end = l_cyl_end + (0.001 * (l_cyl_end - l_cyl_begin));

            difference {
                cylinder {
                    l_cyl_begin
                    l_cyl_end
                    p_minor_radius * 0.9999  // Coincident surface...
                    //open
                    hollow
                    material {
                        texture {
                            t_Traintracks
                            rotate <0, 180, 0>
                            macro_Axis_Remap_Y(l_cyl_begin,
                                l_cyl_end, p_minor_radius)
                        }
                        interior {
                            ior 1.5
                        }

                    }
                }
                cylinder {
                    l_cyl2_begin
                    l_cyl2_end
                    p_minor_radius * (1.0 - p_thickness)
                    //open
                    hollow
                    material {
                        texture {
                            t_Transparent
                        }
                        interior {
                            ior 1
                        }
                    }
                }
            }

            difference {
                cylinder {
                    p_spline(l_i + l_increment - (l_increment * p_join_pct))
                    p_spline(l_i + l_increment + (l_increment * p_join_pct))
                    p_minor_radius + p_thickness * p_minor_radius
                }
                cylinder {
                    p_spline(l_i + l_increment - (l_increment * p_join_pct) - 0.00001)
                    p_spline(l_i + l_increment + (l_increment * p_join_pct) + 0.00001)
                    p_minor_radius
                }
                texture {
                    p_material2
                }
            }
                
            #local l_i = l_i + l_increment;
        #end
    }
#end


#macro macro_Generate_Trainstation_Base(p_height, p_width, p_depth)
    #local l_major_cylinder_radius = p_width * 0.3;
    #local l_minor_cylinder_radius = p_width * 0.1;
    #local l_base_cylinder_bottom = <0, -p_height*0.05, 0>;
    #local l_base_cylinder_top = <0, p_height, 0>;

    union {
        cylinder {
            l_base_cylinder_bottom,
            l_base_cylinder_top,
            l_major_cylinder_radius
        }
        cylinder {
            l_base_cylinder_bottom + <l_major_cylinder_radius, 0, 0>,
            l_base_cylinder_top + <l_major_cylinder_radius, 0, 0>,
            l_minor_cylinder_radius
        }
        cylinder {
            l_base_cylinder_bottom - <l_major_cylinder_radius, 0, 0>,
            l_base_cylinder_top - <l_major_cylinder_radius, 0, 0>,
            l_minor_cylinder_radius
        }
    }
#end


#macro macro_Generate_Trainstation_Head(p_vector_origin, p_vector_out, p_vector_in, p_vector_orientation, p_depth)

    union {
        sphere {
            <0, 0, 0>,
            p_depth
            texture {
                t_Trainstop_Head_Sphere
            }
            translate l_vector_origin
            
        }
        #if (!VZero(p_vector_out))
            cylinder {
                <0, 0, 0>, <0, 1, 0>, 1
                texture {
                    t_Trainstop_Head_Tube
                }
                macro_Axis_Remap_Y(p_vector_origin,
                    p_vector_origin + p_vector_out, p_depth / 2)
            }
        #end
        #if (!VZero(p_vector_in))
            cylinder {
                <0, 0, 0>, <0, 1, 0>, 1
                texture {
                    t_Trainstop_Head_Tube
                }
                macro_Axis_Remap_Y(p_vector_origin,
                    p_vector_origin + p_vector_in, p_depth / 2)
            }
        #end
    }
#end


#macro macro_Generate_Traintracks(p_spline, p_thickness)

    #local l_track_segments = 200;
    
    #local l_i = 0;
    
    #if (g_lod_tube > 0)
        macro_Sectioned_Spline_Lattice(p_spline, p_thickness, 0.01, l_track_segments, 0.025, t_Traintracks, t_Girder)

        // Add the train carriage
        #if (g_lod_carriage > 0)
            macro_Train_Carriage(p_spline(0.084), p_spline(0.090), p_thickness * 0.9)
        #end
    #end
#end

#macro macro_Generate_Tube(p_terrain, p_segments, p_thickness)

    #local l_hit_array = array[g_num_stations];
    #local l_train_station_height = 1.5;
    #local l_train_station_width = 0.05;
    #local l_train_station_depth = 0.06;
    #local l_train_station_tube_height = l_train_station_height - p_thickness * 1.5;

    #local l_i = 0;
    #while (l_i < g_num_stations)
        #local l_vec = vaxis_rotate(<0, -1, 0>, <1, 0, 0>, g_tube_station_coordinates[l_i][0]);
        #local l_hit = trace (p_terrain, <g_tube_station_coordinates[l_i][1], 0, 0>, l_vec);
        #local l_hit_array[l_i] = l_hit;

        #if (VZero(l_hit))
            #error "Building missed target!"
        #end

        #local l_i = l_i + 1;
    #end

    #local l_trainspline = spline {
        natural_spline
        0.00, <-1.6, -19.0, -4>,
        0.08, <-1.57, -19.47, 0>,
        0.10, <-1.6, -19.3, 1>,
        #local l_i = 0;
        #while (l_i < g_num_stations)
            #local l_spline_t = 0.2 + 0.8 * (l_i / (g_num_stations - 1));
            #local l_add_vector = vnormalize(l_hit_array[l_i] * <0, -1, -1>);

            l_spline_t, l_hit_array[l_i] + (l_add_vector * l_train_station_tube_height),

            #local l_i = l_i + 1;
        #end
    }

    macro_Generate_Traintracks(l_trainspline, p_thickness)

    #local l_spline_exit_offset = 0.005;
    #local l_i = 0;
    #while (l_i < g_num_stations)

        //
        // Build the train station pylon
        //
        object {
            macro_Generate_Trainstation_Base(l_train_station_height, l_train_station_width, l_train_station_depth)
            texture {
                t_Trainstop_Base
            }
            scale 1
            rotate <g_tube_station_coordinates[l_i][0], 0, 0>
            translate l_hit_array[l_i]
        }

        //
        // Build the train station platform
        //
        #local l_spline_t = 0.2 + 0.8 * (l_i / (g_num_stations - 1));
        #local l_vector_origin = l_trainspline(l_spline_t);
        #local l_vector_out = l_trainspline(l_spline_t + l_spline_exit_offset) - l_vector_origin;
        #local l_vector_in = l_trainspline(l_spline_t - l_spline_exit_offset) - l_vector_origin;
        #local l_vector_orientation = vnormalize(l_hit_array[l_i] * <0, 1, 1>);
        object {
            macro_Generate_Trainstation_Head(l_vector_origin, l_vector_out, l_vector_in, l_vector_orientation, l_train_station_depth)
        }
        #local l_i = l_i + 1;
    #end
#end
