#declare g_foliage_seed = seed(0);

#macro macro_Generate_Foliage(p_max_xdist, p_from_xrot, p_to_xrot,
            p_element_size, p_element_count, p_max_slope, p_max_distance,
            p_obj_target, p_obj_avoid, p_leaf_textures)
    #if (g_lod_foliage > 0)
        #local l_i = 0;
        #local l_actual_foliage = 0;
    
        #local l_from_xrot = p_from_xrot;
        #while (l_from_xrot < 0)
            #local l_from_xrot = l_from_xrot + 360;
        #end
        #local l_from_xrot = mod(l_from_xrot, 360);
    
        #local l_to_xrot = p_to_xrot;
        #while (l_to_xrot < 0)
            #local l_to_xrot = l_to_xrot + 360;
        #end
        #local l_to_xrot = mod(l_to_xrot, 360);
    
        #while (l_to_xrot < l_from_xrot)
            #local l_to_xrot = 360 + l_to_xrot;
        #end
    
    
        #while (l_i < p_element_count)
            #local l_xoffset = m_RandRangeS(-p_max_xdist, p_max_xdist,
                        g_foliage_seed);
            #local l_xrotate = m_RandRangeS(l_from_xrot, l_to_xrot,
                        g_foliage_seed);
            #local l_element_size = m_RandRangeS(p_element_size, p_element_size * 3, g_foliage_seed);
    
            #local l_target_origin = <l_xoffset, 0, 0>;
            #local l_target_vector = <0, -cos(radians(l_xrotate)),
                        sin(radians(l_xrotate))>;
            #local l_target_normal = <0, 0, 0>;
            #local l_target_hit = trace(p_obj_target, l_target_origin,
                                    l_target_vector, l_target_normal);
            #local l_obstacle_normal = <0, 0, 0>;
            #local l_obstacle_hit = trace(p_obj_avoid, l_target_origin,
                                    l_target_vector, l_obstacle_normal);
    
            // Figure out if it hit something
            #if (!VZero(l_target_normal))
                #local l_target_distance = vlength(l_target_hit - l_target_origin);
                // Check to see if it's not beyond the maximum distance
                #if (l_target_distance < p_max_distance)
                    // Check to see if the angle is not beyond maximum
                    #local l_temp_max_slope = m_RandRangeS(p_max_slope,
                            p_max_slope * 2, g_foliage_seed);
                    #if (VAngleD(-l_target_vector, l_target_normal) < l_temp_max_slope)
                        #if (!VZero(l_obstacle_normal))
                            #local l_obstacle_distance = vlength(l_obstacle_hit - l_target_origin);
                            #if (l_target_distance < l_obstacle_distance)
                                #local l_actual_foliage = l_actual_foliage + 1;
                                macro_Generate_Foliage_Object(p_obj_target,
                                    p_obj_avoid,
                                    l_element_size, l_target_hit,
                                    l_target_normal, p_leaf_textures)
                            #end
                        #else
                            #local l_actual_foliage = l_actual_foliage + 1;
                            macro_Generate_Foliage_Object(p_obj_target,
                                p_obj_avoid,
                                l_element_size, l_target_hit, l_target_normal,
                                p_leaf_textures)
                        #end
                    #end
                #end
            #end

            #local l_i = l_i + 1;
        #end
        #debug "Total regular trees: "
        #debug str(l_actual_foliage, 0, 0)
        #debug "\n"
    #end
#end

#macro macro_Generate_Foliage_Object(p_obj_ground, p_obj_avoid, p_element_size, p_target_hit,
            p_target_normal, p_leaf_textures)

    // Build the trunk
    #local l_trunk_top = p_target_hit +
        m_RandRangeS(p_element_size/6, p_element_size/2, g_foliage_seed) *
        (vnormalize(p_target_hit * -1) + vnormalize(p_target_normal)) / 2;
    #local l_trunk_bottom = p_target_hit + p_element_size * 0.02 * y;
    #local l_trunk_radius_base = p_element_size * 0.01;
    #local l_trunk_radius_base = m_RandRangeS(l_trunk_radius_base,
        2 * l_trunk_radius_base, g_foliage_seed);
    #local l_trunk_radius_top = l_trunk_radius_base * 0.75;
    union {
        cone {
            l_trunk_bottom,
            l_trunk_radius_base,
            l_trunk_top,
            l_trunk_radius_top
            no_shadow
        }
        sphere {
            l_trunk_top,
            l_trunk_radius_top
            no_shadow
        }
        texture {
            t_Tree_Trunk
            scale p_element_size
        }
    }

    // Build the leaves
    #local l_leaf_texture = Rand_Array_Item(p_leaf_textures, g_foliage_seed)
    #local l_leaves = int(m_RandRangeS(4, 6, g_foliage_seed));
    #local l_avg_angle = 360 / l_leaves;
    #local l_current_angle = 360 * rand(g_foliage_seed);
    #local l_avg_length = m_RandRangeS(p_element_size * 0.75, p_element_size * 1.25, g_foliage_seed);

    #local l_trunk_vector = l_trunk_top - l_trunk_bottom;
    #local l_leaf_vector = VProject_Plane(l_trunk_vector + <0, 0, 1>,
                l_trunk_vector);
    #local l_leaf_vector = vnormalize(l_leaf_vector);

    #while (l_leaves > 0)
        #local l_local_leaf_vector = l_leaf_vector;
        #local l_local_leaf_vector = l_local_leaf_vector - m_RandRangeS(l_avg_length * 0.1, l_avg_length * 0.2, g_foliage_seed);
        #local l_actual_length = m_RandRangeS(l_avg_length * 0.9, l_avg_length * 1.1, g_foliage_seed);
        //
        // Make sure that the leaf is not going to poke through the wall
        // of an obstacle or the ground
        //
        #local l_obstacle_normal = <0, 0, 0>;
        #local l_obstacle_hit = trace(p_obj_avoid, l_trunk_top,
                vaxis_rotate(l_local_leaf_vector, l_trunk_vector,
                l_current_angle + 30), l_obstacle_normal);
        #if (!VZero(l_obstacle_hit))
            #if (vlength(l_obstacle_hit - l_trunk_top) < l_actual_length)
                // We hit the obstacle, shorten leaf accordingly
                #local l_actual_length = vlength(l_obstacle_hit - l_trunk_top);
            #end
        #end
        #local l_obstacle_normal = <0, 0, 0>;
        #local l_obstacle_hit = trace(p_obj_ground, l_trunk_top,
                vaxis_rotate(l_local_leaf_vector, l_trunk_vector,
                l_current_angle + 30), l_obstacle_normal);
        #if (!VZero(l_obstacle_hit))
            #if (vlength(l_obstacle_hit - l_trunk_top) < l_actual_length)
                // We hit the obstacle, shorten leaf accordingly
                #local l_actual_length = vlength(l_obstacle_hit - l_trunk_top);
            #end
        #end
        
        //#local l_base = vnormalize(l_local_leaf_vector) * l_trunk_radius_top;
        //#local l_base = vaxis_rotate(l_base, l_trunk_vector, l_current_angle + 30);
        #local l_base = <0, 0, 0>;
        #local l_left_corner = l_local_leaf_vector * l_actual_length * 0.5;
        #local l_left_corner = vaxis_rotate(l_left_corner, l_trunk_vector, l_current_angle);
        #local l_tip = l_local_leaf_vector * l_actual_length;
        #local l_tip = vaxis_rotate(l_tip, l_trunk_vector, l_current_angle + 30);
        #local l_right_corner = l_local_leaf_vector * l_actual_length * 0.5;
        #local l_right_corner = vaxis_rotate(l_right_corner, l_trunk_vector, l_current_angle + 60);

        // Add some 'turbulence'
        #local l_left_corner = l_left_corner + m_RandRangeS(l_avg_length * 0.05, l_avg_length * 0.1, g_foliage_seed);
        #local l_right_corner = l_right_corner + m_RandRangeS(l_avg_length * 0.05, l_avg_length * 0.1, g_foliage_seed);
        #local l_top_normal = l_trunk_vector;
        #local l_tip_normal = l_trunk_top * 5 + l_tip - l_trunk_bottom * 5;
        #local l_left_normal = l_trunk_top * 5 + l_left_corner - l_trunk_bottom * 5;
        #local l_right_normal = l_trunk_top * 5 + l_right_corner - l_trunk_bottom * 5;

        mesh {
            smooth_triangle {
                l_trunk_top + l_base, l_top_normal,
                l_trunk_top + l_left_corner, l_left_normal,
                l_trunk_top + l_right_corner, l_right_normal
                uv_vectors <0, 0>, <0, 1>, <1, 0>
                texture {
                    l_leaf_texture
                }
            }

            smooth_triangle {
                l_trunk_top + l_left_corner, l_left_normal,
                l_trunk_top + l_right_corner, l_right_normal,
                l_trunk_top + l_tip, l_tip_normal
                uv_vectors <0, 1>, <1, 0>, <1, 1>
                texture {
                    l_leaf_texture
                }
            }
            double_illuminate
            no_shadow
        }

        #local l_current_angle = l_current_angle + l_avg_angle;
        #local l_leaves = l_leaves - 1;
    #end

#end
