//  
// Lost In Space
//


#declare g_lod_fog = 1;
#declare g_lod_river = 1;
#declare g_lod_river_media = 1;
#declare g_lod_plasma = 1;
#declare g_lod_detailmesh = 1;
#declare g_lod_cables = 1;
#declare g_lod_tube = 1;
#declare g_lod_carriage = 1;
#declare g_lod_lattice = 1;
#declare g_lod_arealight = 1;
#declare g_lod_filllight = 1;
#declare g_lod_detailtextures = 1;
#declare g_lod_transparency = 1;
#declare g_lod_foliage = 1;
#declare g_lod_structures = 1;
#declare g_lod_focalblur = 0;   // Normally, focal blur and radiosity are off, but if you
#declare g_lod_radiosity = 0;   // have access to a supercomputer or two, go right ahead


#include "arrays.inc"
#include "colors.inc"
#include "functions.inc"
#include "math.inc"
#include "transforms.inc"

#include "ss_utilities.inc"
#include "ss_textures.inc"
#include "ss_cables.inc"
#include "ss_structures.inc"
#include "ss_shell.inc"
#include "ss_landscape.inc"
#include "ss_carriage.inc"
#include "ss_ducting.inc"
#include "ss_powergenerator.inc"
#include "ss_foliage.inc"


//
// Set up some defaults
//

global_settings {
    max_trace_level 16
    adc_bailout 0.02
    #if (g_lod_radiosity > 0)
        radiosity {
        }
    #end
}

#declare g_observer_position =  <-1.5, -19.5, -0.00>;
#declare g_observer_target = <0, -10, 100>;

#macro macro_Generate_Station(p_radius_major, p_radius_minor)
    light_group {
        #local l_depth = p_radius_minor * 0.5;
        macro_Generate_Shell_Cover(p_radius_major, p_radius_minor, 80, 10, 0.02)
        macro_Generate_Shell_Base(p_radius_major, p_radius_minor, l_depth)

        #local lo_Station_Terrain = object {
            macro_Generate_Terrain(p_radius_major, p_radius_minor)
        }

        macro_Generate_Tube(lo_Station_Terrain, 200, 0.02)

        #if (g_lod_structures > 0)
            #local lo_Station_Foliage_Obstacles = object {
                macro_Generate_Structures(lo_Station_Terrain, 10000, p_radius_minor * 1.3, p_radius_major * 1.03, 25, 0.8, 0.5)
            }

            lo_Station_Foliage_Obstacles

            macro_Generate_Foliage(p_radius_minor * 2, -5, 120, 0.05, 120000, 15, p_radius_major + (p_radius_minor * 0.4), lo_Station_Terrain, lo_Station_Foliage_Obstacles, t_Leaf_Array)

            #undef lo_Station_Foliage_Obstacles
        #end

        #undef lo_Station_Terrain

        macro_Generate_Plasma_Rail(p_radius_major, p_radius_minor, p_radius_minor / 10, 50)
        macro_Generate_Landscape(p_radius_major, p_radius_minor, l_depth)

    }

    light_group {
        light_source {
            <0, 0, 0>
            color <1, 1, 1>
        }
        macro_Generate_Shell_Fog(p_radius_major, p_radius_minor, l_depth)
        macro_Generate_River(p_radius_major, p_radius_minor, l_depth)
    }
#end


sky_sphere {
    pigment {
    	p_Star_Map
    }
}

macro_Generate_Station(20, 3)

camera {
    location g_observer_position
    look_at g_observer_target

    angle 110       // Wide angle
    right x * 2     // 2:1 aspect ratio

    #if (g_lod_focalblur > 0)
        aperture 0.0001
        blur_samples 20
        focal_point <0, 0, -2>
    #end
}
