
#macro macro_Sectioned_Plasma_Rail(p_major_radius, p_minor_radius, p_thickness, p_segments, p_join, p_light_source, p_light_begin, p_light_end, p_material1)

    #local l_increment = 2 * pi / p_segments;
    #local l_theta = 0;
    #local l_area_lights = 0;

    #while (l_theta < (2 * pi))
        #local l_x1 = p_major_radius * cos(l_theta);
        #local l_y1 = p_major_radius * sin(l_theta);
        #local l_z1 = 0;
        #local l_x2 = p_major_radius * cos(l_theta + l_increment);
        #local l_y2 = p_major_radius * sin(l_theta + l_increment);
        #local l_z2 = 0;

        cylinder {
            <0, 0, 0>
            <0, 1, 0>
            1
            hollow
            no_shadow
            material {
                p_material1
            }
            macro_Axis_Remap_Y(<l_x1, l_y1, l_z1>, <l_x2, l_y2, l_z2>, p_minor_radius)
            rotate <0, 90, 0>
        }

        #if (g_lod_filllight > 0)
            #if ((degrees(l_theta) <= p_light_end) &
                    ((degrees(l_theta)) >= p_light_begin))
                #local l_area_lights = l_area_lights + 1;
                object {
                    p_light_source
                    translate <0, 0.5, 0>
                    macro_Axis_Remap_Y(<l_x1, l_y1, l_z1>, <l_x2, l_y2, l_z2>, p_minor_radius)
                    rotate <0, 90, 0>
                }
            #end
        #end
        
        object {
            p_join
            rotate <0, 0, degrees(l_theta)>
            translate <l_x1, l_y1, l_z1>
            rotate <0, 90, 0>
        }
        
        #local l_theta = l_theta + l_increment;
    #end

    #debug "Total area lights: "
    #debug str(l_area_lights, 0, 0)
    #debug "\n"

#end

#macro macro_Generate_Plasma_Rail(p_major_radius, p_minor_radius, p_rail_radius, p_rail_segments)

    #if (g_lod_plasma > 0)
        #local o_Section_Join = object {
            torus {
                p_rail_radius,
                p_rail_radius / 20
                material {
                	macro_t_Generator_Hub(p_rail_radius)
                }
                scale <1, 15, 1>
                no_shadow
            }
        }
        #if (g_lod_arealight > 0)
            #local o_Section_Light = object {
                light_source {
                    <0, 0.5, 0>
                    color <1, 1, 1> * 2
                    area_light
                    <0, 0.9, 0>, <0.01, 0, 0>, 9, 1
                    adaptive 0
                    fade_power 2
                    fade_distance 1
                }
            }
        #else
            #local o_Section_Light = object {
                light_source {
                    <0, 0, 0>
                    color <1, 1, 1> * 2
                    fade_power 2
                    fade_distance 1
                }
            }
        #end
        #local m_Mat1 = material {
            texture {
                pigment {
                    rgbt <1, 1, 1, 1>
                }
            }
            interior {
                media {
                    emission 10
                    density {
                        crackle
                        color_map {
                            [0.01 rgb <1, 1, 1>]
                            [0.05 rgb <0, 0, 1>]
                            [0.1 rgb <0, 0, 0>]
                            [1.0 rgb <0, 0, 0>]
                        }
                        scale 0.2
                    }
                }
                media {
                    emission 1
                    density {
                        cylindrical
                        color_map {
                            [0.0 rgb <0, 0, 0>]
                            [0.05 rgb <0.1, 0.1, 0.25>]
                            [0.4 rgb <1, 1, 1>]
                            [1.0 rgb <1, 1, 1>]
                        }
                    }
                }
            }
        }

        macro_Sectioned_Plasma_Rail(p_major_radius - p_minor_radius + p_rail_radius, p_rail_radius, 0, p_rail_segments, o_Section_Join, o_Section_Light, 150, 300, m_Mat1)
    #end
    #if (g_lod_filllight = 0)
        light_source {
            <0, 0, 0>
            color <1, 1, 1>
        }
    #end
#end
