

#macro macro_Generate_Shell_Cover(p_radius_major, p_radius_minor, p_increments_major, p_increments_minor, p_thickness)

    #if (g_lod_lattice > 0)
        #declare new_welds = array [p_increments_major];
        #declare old_welds = array [p_increments_major];
        union {

            #local l_rim_increment = pi / p_increments_minor;
            #local l_theta = -pi / 2;
            #local l_minor_iter = 0;
            #local l_rho_offset = 0;
            #while (l_minor_iter <= p_increments_minor)
                //
                // Build the welds & struts
                //
                #local l_rho = 0;
                #local l_major_iter = 0;
                #local l_maj_increment = 2 * pi / p_increments_major;
                #local l_rho_offset = l_rho_offset - l_maj_increment / 2;
                #while (l_rho < 2 * pi)
                    #local l_x = 2 * p_radius_minor * sin(l_theta);
                    #local l_y = (p_radius_major - p_radius_minor * cos(l_theta)) * cos(l_rho + l_rho_offset);
                    #local l_z = (p_radius_major - p_radius_minor * cos(l_theta)) * sin(l_rho + l_rho_offset);
                    #declare new_welds[l_major_iter] = <l_x, l_y, l_z>;
                    #if (l_minor_iter != 0)
                        #if (l_minor_iter = p_increments_minor)
                            macro_Generate_Pane(l_major_iter, p_increments_major, 1, p_thickness)
                        #else
                            macro_Generate_Pane(l_major_iter, p_increments_major, 0, p_thickness)
                        #end
                    #end

                    #local l_rho = l_rho + l_maj_increment;
                    #local l_major_iter = l_major_iter + 1;
                #end
                #declare old_welds = new_welds;
                #local l_theta = l_theta + l_rim_increment;
                #local l_minor_iter = l_minor_iter + 1;
            #end
            hollow
        }
    #end
#end

#macro macro_Generate_Shell_Base(p_radius_major, p_radius_minor, p_depth)
    //
    // Build the base
    //
    // Build a vertical cylinder, and two horizontal cylinders, and a pane
    // consisting of two triangles.
    //
    #local l_floor_depth = p_radius_major + p_depth;
    #local l_wall_offset = p_radius_minor * 1.75;
    #local l_segments = 200;
    #local l_thickness = 0.1;

    macro_Generate_Shell_Base_Side(l_wall_offset, p_radius_minor * 2, l_floor_depth, p_radius_major, l_segments, l_thickness)
    macro_Generate_Shell_Base_Side(-l_wall_offset, -p_radius_minor * 2, l_floor_depth, p_radius_major, l_segments, l_thickness)

#end

#macro macro_Generate_Shell_Base_Side(p_xoffset1, p_xoffset2, p_radius1, p_radius2, p_segments, p_thickness)
    //
    // Build the base
    //
    // Build a vertical cylinder, and two horizontal cylinders, and a pane
    // consisting of two triangles.
    //
    #local l_theta = 0;
    #local l_increment = 2 * pi / p_segments;
    #local l_i = 0;

    #while (l_i < p_segments)
        #local point_A1 = <p_xoffset1, p_radius1 * cos(l_theta),
                p_radius1 * sin(l_theta)>;
        #local point_B1 = <p_xoffset2, p_radius2 * cos(l_theta),
                p_radius2 * sin(l_theta)>;

        #local l_theta = l_theta + l_increment;

        #local point_A2 = <p_xoffset1, p_radius1 * cos(l_theta),
                p_radius1 * sin(l_theta)>;
        #local point_B2 = <p_xoffset2, p_radius2 * cos(l_theta),
                p_radius2 * sin(l_theta)>;

        union {
            cylinder {
                point_A1,
                point_B1,
                p_thickness / 2
            }
            cylinder {
                point_A1,
                point_A2,
                p_thickness / 2
            }
            cylinder {
                point_B1,
                point_B2,
                p_thickness / 2
            }
            sphere {
                point_A1,
                p_thickness / 2
            }
            sphere {
                point_B1,
                p_thickness / 2
            }
            texture {
                t_Shell_Pane_Strut
            }
        }

        union {
            triangle {
                point_A1,
                point_A2,
                point_B2
            }
            triangle {
                point_A1,
                point_B2,
                point_B1
            }
            texture {
                t_Shell_Obsidian
            }
        }


        #local l_i = l_i + 1;
    #end

#end

#macro macro_Generate_Shell_Fog(p_radius_major, p_radius_minor, p_depth)
    #if (g_lod_fog > 0)
        torus {
            p_radius_major, p_radius_minor * 0.95
            scale <1, 2, 1>
            rotate <0, 0, 90>
            hollow
            no_shadow
            no_reflection
            material {
                m_Shell_Fog
            }
        }
    #end

#end

//
// vc |\   | vd
//    | \Q |
//    | P\ |
// vb |___\| va  <------p_iter
//
#macro macro_Generate_Pane(p_iter, p_max, p_last, p_thickness)
    #local l_va = new_welds[p_iter];
    #local l_vb = old_welds[p_iter];
    #if (p_iter != 0)
        #local l_vc = old_welds[p_iter - 1];
        #local l_vd = new_welds[p_iter - 1];
    #else
        #local l_vc = old_welds[p_max - 1];
    #end

    // Angled Strut 1
    macro_Generate_Strut(l_va, l_vb, p_thickness)
    // Angled Strut 2
    macro_Generate_Strut(l_va, l_vc, p_thickness)

    // Rim Strut
    macro_Generate_Strut(l_vb, l_vc, p_thickness)

    // Last Rim Strut
    #if (p_last != 0)
        // The first last rim strut isn't generated on the first instance
        #if (p_iter != 0)
            macro_Generate_Strut(l_va, l_vd, p_thickness)
        #end
        // Make up the first last rim strut now that we have the missing information
        #if (p_iter = (p_max - 1))
            macro_Generate_Strut(l_va, new_welds[0], p_thickness)
        #end
    #end

    // Glass Pane P
    macro_Generate_Glass(l_va, l_vb, l_vc)

    // If it's not the first iteration then generate pane Q as well
    #if (p_iter != 0)
        macro_Generate_Glass(l_va, l_vc, l_vd)
    #end

    #if (p_iter = (p_max - 1))
        // It's the last iteration, so make up the first Q pane now.
        macro_Generate_Glass(l_va, l_vc, new_welds[0])
    #end
#end

#macro macro_Generate_Strut(p_va, p_vb, p_thickness)
    cylinder {
        p_va, p_vb, p_thickness
        texture {
            t_Shell_Pane_Strut
        }
        hollow
        no_shadow
    }
#end

#macro macro_Generate_Glass(p_va, p_vb, p_vc)
    triangle {
        p_va + <0.0001, 0.0001, 0.0001>, p_vb + <0.0001, 0.0001, 0.0001>, p_vc + <0.0001, 0.0001, 0.0001>
        hollow
        no_shadow
        no_reflection
        texture {
            t_Shell_Pane_Glass_Null
        }
    }
    triangle {
        p_vc, p_vb, p_va
        hollow
        no_shadow
        no_reflection
        texture {
            t_Shell_Pane_Glass
        }
    }
#end

