//
// Defines the buildings that are scattered througout the scene
//

#declare g_Tex_Rand = seed(0);
#declare g_Struct_Rand = seed(2);

#macro macro_Building_Sphere_Section(p_radius, p_high_offset, p_low_offset, p_displacement, p_tex_main)    
    // Create a slice of a sphere by doing an intersection with a box
    object {
    intersection {
            sphere {
                <0, p_radius, 0>, p_radius
            }
            box {
                <-1.01, p_low_offset * p_radius * 2, -1.01>
                <1.01, p_high_offset * p_radius * 2, 1.01>
            }       
            texture {
                p_tex_main
            }
        }
        translate <0, p_displacement - p_radius * 2 * p_low_offset, 0>
    }
#end

#macro macro_Building_StructureA_Head(p_radius, p_base_pct, p_apex_pct, p_segments, p_connector_proportion, p_tex_main, p_tex_window)
    #local l_increment = (p_apex_pct - p_base_pct) / p_segments;
    #local l_i = 0;
    #local l_displacement = 0;
    #local l_actual_size = 2 * p_radius;
    #local l_cyl_displacement = l_increment * p_connector_proportion * l_actual_size;
    
    union {
    
        #while (l_i < p_segments)
            #local l_upper = p_base_pct + (l_i + 1) * l_increment;
            #local l_lower = p_base_pct + l_i * l_increment;

            macro_Building_Sphere_Section(p_radius, l_upper, l_lower, l_displacement, p_tex_main)
            #local l_displacement = l_displacement + l_increment * (2 * p_radius);
        
            #if (l_upper > 0.5)
                #local l_cyl_radius = sqrt(p_radius * p_radius - (l_actual_size *
                        (l_upper - 0.5)) * (l_actual_size * (l_upper - 0.5)));
            #else
                #local l_cyl_radius = sqrt(p_radius * p_radius - (l_actual_size *
                        (0.5 - l_lower - l_increment)) * (l_actual_size *
                        (0.5 - l_lower - l_increment)));
            #end
        
            #local l_i = l_i + 1;
        
            #if (l_i != p_segments)
                cylinder {
                    <0, l_displacement, 0>
                    <0, l_displacement + l_cyl_displacement, 0>
                    l_cyl_radius //* 0.95
                    texture {
                        p_tex_window
                    }
                }

                #local l_displacement = l_displacement + l_cyl_displacement;
            #end    
        #end 
    }
#end

#macro macro_Building_StructureA_Base(p_thickness, p_radius, p_height, p_tex_main)
    object {
        intersection {
            sor {
                8,
                <0.9, -0.5>
                <0.9,  0.0>
                <0.9,  0.2500>
                <0.8,  0.4375>
                <0.6,  0.6250>
                <0.5,  0.8125>
                <0.5,  1.0>
                <0.5,  1.5>   
            }
            box {
                <-p_thickness/2, 1, 1>
                <p_thickness/2, 0, -1>
            }
            texture {
                p_tex_main
            }
        }
        scale <p_radius, p_height, p_radius>
    }
#end

#macro macro_Building_StructureA(p_head_radius, p_head_base_pct, p_head_apex_pct, p_head_segments, p_head_connector_proportion, p_base_radius, p_base_thickness, p_base_height, p_tex_main, p_tex_window)
    union {
        object {
            macro_Building_StructureA_Head(p_head_radius / 2, p_head_base_pct, p_head_apex_pct,
                p_head_segments, p_head_connector_proportion, p_tex_main, p_tex_window)
            translate <0, p_base_height, 0>
        }
        object {
            macro_Building_StructureA_Base(p_base_thickness, p_base_radius, p_base_height, p_tex_main)
        }    
        object {
            macro_Building_StructureA_Base(p_base_thickness, p_base_radius, p_base_height, p_tex_main)
            rotate <0, 120, 0>
        }    
        object {
            macro_Building_StructureA_Base(p_base_thickness, p_base_radius, p_base_height, p_tex_main)
            rotate <0, -120, 0>
        }
    }
#end

#macro macro_Building_StructureD(p_width, p_base_height, p_top_height, p_cyl_depth, p_tex_base, p_tex_top, p_tex_roof)
    //
    // Build this building using a box for the base, a box for the top,
    // differenced with cylinders.
    //
    // Do it in the coordinates <-1, 0, -1>,<1, 1, 1> to make texturing simple
    // then scale to the right proportions
    //

    #local l_base_height = p_base_height / (p_base_height + p_top_height);
    #local l_cyl_depth = p_cyl_depth / p_width;
    #local l_theta = atan(1 / p_cyl_depth);
    #local l_hyp = sqrt((1 * 1) + (l_cyl_depth * l_cyl_depth));
    #local l_radius = l_hyp * sin(l_theta) / sin(180 - 2 * l_theta);
    #local l_offset = 1 + l_radius - l_cyl_depth;

    #local l_tex_top = texture {
        p_tex_top
        scale (1 - l_base_height)
        translate <0, l_base_height, 0>
    }

    union {
/*
        box {
            <-1, 0, -1>,
            <1, l_base_height, 1>
            texture {
                p_tex_base
            }
        }
*/
        difference {
            box {
                <-1, -l_base_height, -1>,
                <1, l_base_height, 1>
            }
            cylinder {
                <-l_offset, -l_base_height -0.0001, 0>
                <-l_offset, l_base_height + 0.0001, 0>
                l_radius
            }
            cylinder {
                <l_offset, -l_base_height -0.0001, 0>
                <l_offset, l_base_height + 0.0001, 0>
                l_radius
            }
            cylinder {
                <0, -l_base_height -0.0001, -l_offset>
                <0, l_base_height + 0.0001, -l_offset>
                l_radius
            }
            cylinder {
                <0, -l_base_height -0.0001, l_offset>
                <0, l_base_height + 0.0001, l_offset>
                l_radius
            }
            texture {
                p_tex_base
            }
        }
        difference {
            box {
                <-1, l_base_height, -1>,
                <1, 1, 1>
            }
            cylinder {
                <-l_offset, l_base_height-0.0001, 0>
                <-l_offset, 1.0001, 0>
                l_radius
            }
            cylinder {
                <l_offset, l_base_height-0.0001, 0>
                <l_offset, 1.0001, 0>
                l_radius
            }
            cylinder {
                <0, l_base_height-0.0001, -l_offset>
                <0, 1.0001, -l_offset>
                l_radius
            }
            cylinder {
                <0, l_base_height-0.0001, l_offset>
                <0, 1.0001, l_offset>
                l_radius
            }
            texture {
                l_tex_top
            }
        }
        scale <p_width, p_base_height + p_top_height, p_width>
    }
#end

#macro macro_Building_StructureC(p_width, p_breadth, p_base_height, p_top_height, p_tex_base, p_tex_top, p_tex_roof)
    //
    // Build this building using a box for the base, a cylinder for the top,
    // intersected with a box.
    //
    // Do it in the coordinates <-1, 0, -1>,<1, 1, 1> to make texturing simple
    // the scale to the right proportions
    //

    #if (p_breadth >= p_width)
        #error "Breadth cannot be greater than or equal to width here\n"
    #end

    #local l_base_width = p_breadth / p_width;
    #local l_base_height = p_base_height / (p_base_height + p_top_height);

    #local l_tex_top = texture {
        p_tex_top
        scale (1 - l_base_height)
        translate <0, l_base_height, 0>
    }

    union {
/*
        box {
            <-1, 0, -l_base_width>,
            <1, l_base_height, l_base_width>
            texture {
                p_tex_base
            }
        }
*/
        intersection {
            cylinder {
                <0, -l_base_height, 0>,
                <0, l_base_height, 0>,
                1
            }
            box {
                <-1, -l_base_height - 0.0001, -l_base_width>,
                <1, l_base_height + 0.0001, l_base_width>
            }
            texture {
                p_tex_base
            }
        }
        intersection {
            cylinder {
                <0, l_base_height, 0>,
                <0, 1, 0>,
                1
            }
            box {
                <-1, l_base_height - 0.0001, -l_base_width>,
                <1, 1.0001, l_base_width>
            }
            texture {
                l_tex_top
            }
        }
        scale <p_width, p_base_height + p_top_height, p_width>
    }
#end


#macro macro_Building_StructureB(p_base_width, p_base_breadth, p_top_width, p_top_breadth, p_base_height, p_top_height, p_tex_base, p_tex_top, p_tex_roof)
    //
    // Build this building using a box for the base, and triangles for the top
    // and sides.
    //

    #local l_x_base = 1;
    #local l_x_top = p_top_width / p_base_width;
    #local l_z_base = p_base_breadth / p_base_width;
    #local l_z_top = p_top_breadth / p_base_width;
    #local l_base_height = p_base_height / (p_base_height + p_top_height);
    //#local l_x_base = p_base_width / 2;
    //#local l_x_top = p_top_width / 2;
    //#local l_z_base = p_base_breadth / 2;
    //#local l_z_top = p_top_breadth / 2;

    #local l_tex_top = texture {
        p_tex_top
        scale (1 - l_base_height)
        translate <0, l_base_height, 0>
    }

    union {
        box {
            <-l_x_base, -l_base_height, -l_z_base>,
            <l_x_base, l_base_height, l_z_base>
            texture {
                p_tex_base
            }
        }

        mesh {
            // Front
            triangle {
                <-l_x_base, l_base_height, -l_z_base>,
                <l_x_top, 1, -l_z_top>,
                <l_x_base, l_base_height, -l_z_base>
                texture {
                    l_tex_top
                }
            }
            triangle {
                <-l_x_base, l_base_height, -l_z_base>,
                <-l_x_top, 1, -l_z_top>,
                <l_x_top, 1, -l_z_top>
                texture {
                    l_tex_top
                }
            }
            // Right
            triangle {
                <l_x_base, l_base_height, -l_z_base>,
                <l_x_top, 1, l_z_top>,
                <l_x_base, l_base_height, l_z_base>
                texture {
                    l_tex_top
                }
            }
            triangle {
                <l_x_base, l_base_height, -l_z_base>,
                <l_x_top, 1, -l_z_top>,
                <l_x_top, 1, l_z_top>
                texture {
                    l_tex_top
                }
            }
            // Back
            triangle {
                <l_x_base, l_base_height, l_z_base>,
                <-l_x_top, 1, l_z_top>,
                <-l_x_base, l_base_height, l_z_base>
                texture {
                    l_tex_top
                }
            }
            triangle {
                <l_x_base, l_base_height, l_z_base>,
                <l_x_top, 1, l_z_top>,
                <-l_x_top, 1, l_z_top>
                texture {
                    l_tex_top
                }
            }
            // Left
            triangle {
                <-l_x_base, l_base_height, l_z_base>,
                <-l_x_top, 1, -l_z_top>,
                <-l_x_base, l_base_height, -l_z_base>
                texture {
                    l_tex_top
                }
            }
            triangle {
                <-l_x_base, l_base_height, l_z_base>,
                <-l_x_top, 1, l_z_top>,
                <-l_x_top, 1, -l_z_top>
                texture {
                    l_tex_top
                }
            }
            // Top
            triangle {
                <-l_x_top, 1, -l_z_top>,
                <l_x_top, 1, l_z_top>,
                <l_x_top, 1, -l_z_top>
                texture {
                    p_tex_roof
                }
            }
            triangle {
                <-l_x_top, 1, -l_z_top>,
                <-l_x_top, 1, l_z_top>,
                <l_x_top, 1, l_z_top>
                texture {
                    p_tex_roof
                }
            }
            // Base
            triangle {
                <-l_x_base, l_base_height, l_z_base>,
                <l_x_base, l_base_height, -l_z_base>,
                <l_x_base, l_base_height, l_z_base>
            }
            triangle {
                <-l_x_base, l_base_height, l_z_base>,
                <-l_x_base, l_base_height, -l_z_base>,
                <l_x_base, l_base_height, -l_z_base>
            }
        }
        scale <p_base_width, p_base_height + p_top_height, p_base_width>
    }
#end

#macro macro_Structure_Proximity(p_hit, p_structure_actual, p_structure_array, p_threshold)
    #local l_sp_iter = 0;
    #local l_sp_ret = 1;
    #while ((l_sp_iter < p_structure_actual) & (l_sp_ret > 0))
        #if (vlength(p_hit - p_structure_array[l_sp_iter]) < p_threshold)
            #local l_sp_ret = 0;
        #end
        #local l_sp_iter = l_sp_iter + 1;
    #end

    l_sp_ret
#end


#macro macro_Generate_Structures(p_target, p_num_buildings, p_xmax, p_max_distance, p_max_slope, p_threshold, p_scale)

    // Algorithm:
    // Pick a random offset on the X axis (axis of rotation of station)
    // Pick a random angle about the X axis
    // Shoot a ray in that direction
    // If the ray hits the target object, check the slope and distance there.
    // If the slope is too steep, discard.
    // If the distance is too far (eg, in the river), discard
    // Check distance from all other hits.  If too close, discard.
    // Randomly select a building type.
    // Texture it.
    // Scale it appropriately
    // Rotate it about it's Y axis to get a bit of variety
    // Rotate it about it's X axis by the previously selected angle.
    // Translate it to the hit point.
    // Translate it slightly further into the ground so that it goes in well.
    // Add structure to union of existing structures.
    // Repeat n times.

    #if (g_lod_structures > 0)
        #local l_structure_count = 0;
        #local l_structure_actual = 0;
        #local l_structure_array = array[p_num_buildings];

        #while (l_structure_count < p_num_buildings)
            #local l_xoffset = m_RandRangeS(-p_xmax, p_xmax, g_Struct_Rand);
            #local l_xrotate = m_RandRangeS(0, 360, g_Struct_Rand);

            // Shoot a ray at the ground
            #local l_target_origin = <l_xoffset, 0, 0>;
            #local l_target_vector = <0, -cos(radians(l_xrotate)),
                        sin(radians(l_xrotate))>;
            #local l_target_normal = <0, 0, 0>;
            #local l_target_hit = trace(p_target, l_target_origin,
                                    l_target_vector, l_target_normal);
            #local l_target_distance = 999999;
            #if (!VZero(l_target_normal))
                #local l_target_distance =
                    vlength(l_target_hit - l_target_origin);
            #end

            // Check to see if the target distance is too far
            #if (l_target_distance <= p_max_distance)
                // Check to see if the slope is too steep
                #if (VAngleD(-l_target_vector, l_target_normal) < p_max_slope)
                    // Check to see if there are any other buildings too close.
                    #if (macro_Structure_Proximity(l_target_hit,
                            l_structure_actual, l_structure_array,
                            p_threshold) > 0)
                        // Add the coordinates to the array.
                        #local l_structure_array[l_structure_actual] = 
                            l_target_hit;
                        #local l_structure_actual = l_structure_actual + 1;
                    #end
                #end
            #end

            #local l_structure_count = l_structure_count + 1;
        #end

        // Now that we've figured out all of the coordinates,
        // place the buildings.
        
        union {
            #local l_structure_count = 0;
            #while (l_structure_count < l_structure_actual)
                #local l_struct_textures = array[3];
                macro_Calculate_Building_Textures(l_struct_textures, 15, 70)

                object {
                    #switch (m_RandRangeS(0, 8, g_Struct_Rand))
                        #range (0, 2)
                            macro_Building_StructureA(1.0, 0.2, 1, 10, 0.6, 0.5, 0.15, 1, l_struct_textures[1], l_struct_textures[2])
                            #break
                        #range (2, 4)
                            macro_Building_StructureB(0.9, 0.5, 0.4, 0.2, 0.5, 2.2, l_struct_textures[1], l_struct_textures[0], l_struct_textures[1])
                            #break
                        #range (4, 6)
                            macro_Building_StructureC(0.6, 0.5, 0.5, 2.0, l_struct_textures[1], l_struct_textures[0], l_struct_textures[1])
                            #break
                        #range (6, 8)
                            macro_Building_StructureD(0.4, 0.5, 2.2, 0.1, l_struct_textures[1], l_struct_textures[0], l_struct_textures[1])
                            #break
                    #end

                    #local l_hit = l_structure_array[l_structure_count];
                    #local l_angle = VAngleD(<0, -1, 0>, l_hit * <0, 1, 1>);
                    // Scale to requested size
                    scale p_scale
                    // Add y axis randomness
                    rotate <0, m_RandRangeS(-15, 15, g_Struct_Rand), 0>
                    // Rotate correct angle in X axis
                    rotate <-l_angle, 0, 0>
                    // Translate to hit location
                    translate l_hit
                    // Sink into ground a bit
                    translate vnormalize(l_hit * <0, 1, 1>) * 0.3 * p_scale
                }
                #local l_structure_count = l_structure_count + 1;
            #end
        }
        #debug "Total regular buildings: "
        #debug str(l_structure_count, 0, 0)
        #debug "\n"
        #undef l_structure_array
    #end
#end
