
#declare g_util_default_seed = seed(0);

#macro m_RandRange(p_from, p_to)
    (rand(g_util_default_seed) * (p_to - p_from) * 0.99999999 + p_from)
#end

#macro m_RandRangeS(p_from, p_to, p_seed)
    (rand(p_seed) * (p_to - p_from) * 0.99999999 + p_from)
#end

#macro DEBUG_VECTOR(p_name, p_vec)

    #debug p_name
    #debug ": <"
    #debug str(p_vec.x, 0, 4)
    #debug ", "
    #debug str(p_vec.y, 0, 4)
    #debug ", "
    #debug str(p_vec.z, 0, 4)
    #debug ">\n"

#end

#macro macro_Axis_Remap(p_A, p_B, p_rad)
    // This could get tricky.  What I want to do is do a remapping from
    // cylinder {<-1, 0, 0>, <1, 0, 0>, 1} to
    // cylinder {p_A, p_B, p_rad}.

    #local l_reorient = p_B - p_A;
    #local l_scale = <vlength(l_reorient) / 2, p_rad, p_rad>;
    #local l_translate = (p_A + p_B) / 2;
    #local l_commonaxis = vcross(<1, 0, 0>, l_reorient);
    #local l_angle = VRotationD(<1, 0, 0>, l_reorient, l_commonaxis);

    // It's not good enough just to remap the cylinder, we also need to make
    // sure that it's in the right radial rotation.  As we want the bottom of
    // the cylinder to map to the <0, Y, Z>, we make a vector <0, -1, 0>, then
    // remap it to see where it ends up.  We then figure out how far, and in 
    // what direction it needs to be rotated in order to end up where we would
    // want it, then apply that transformation to the object prior to
    // translation.
    #local l_aux_actual = <0, -1, 0> * l_scale;
    #local l_aux_actual = vaxis_rotate(l_aux_actual, l_commonaxis, l_angle);
    #local l_aux_desired = l_translate * <0, 1, 1>;
    #local l_aux_angle = VRotationD(l_aux_actual, l_aux_desired, l_reorient);

    scale l_scale
    rotate <l_aux_angle, 0, 0>
    Axis_Rotate_Trans(l_commonaxis, l_angle)
    translate l_translate
#end

#macro macro_Axis_Remap_Y(p_A, p_B, p_rad)
    // Similar to macro_Axis_Remap, but remaps a cylinder that exists on
    // cylinder {<0, 0, 0>, <0, 1, 0>, 1} for ease of use with the
    // cylindrical, radial, etc patterns.

    #local l_reorient = p_B - p_A;
    #local l_scale = <p_rad, vlength(l_reorient), p_rad>;
    #local l_translate = p_A;
    #local l_commonaxis = vcross(<0, 1, 0>, l_reorient);
    #local l_angle = VRotationD(<0, 1, 0>, l_reorient, l_commonaxis);

    scale l_scale
    Axis_Rotate_Trans(l_commonaxis, l_angle)
    translate l_translate
#end

#macro macro_Axis_Remap_Point(p_V, p_A, p_B, p_rad)
    // Do the same thing as the macro_Axis_Remap macro but for a single vector

    #local l_reorient = p_B - p_A;
    #local l_scale = <vlength(l_reorient) / 2, p_rad, p_rad>;
    #local l_translate = (p_A + p_B) / 2;
    #local l_commonaxis = vcross(<1, 0, 0>, l_reorient);
    #local l_angle = VRotationD(<1, 0, 0>, l_reorient, l_commonaxis);

    // It's not good enough just to remap the cylinder, we also need to make
    // sure that it's in the right radial rotation.  As we want the bottom of
    // the cylinder to map to the <0, Y, Z>, we make a vector <0, -1, 0>, then
    // remap it to see where it ends up.  We then figure out how far, and in 
    // what direction it needs to be rotated in order to end up where we would
    // want it, then apply that transformation to the object prior to
    // translation.
    #local l_aux_actual = <0, -1, 0> * l_scale;
    #local l_aux_actual = vaxis_rotate(l_aux_actual, l_commonaxis, l_angle);
    #local l_aux_desired = l_translate * <0, 1, 1>;
    #local l_aux_angle = VRotationD(l_aux_actual, l_aux_desired, l_reorient);

    #local l_newV = p_V * l_scale;
    #local l_newV = vaxis_rotate(l_newV, <1, 0, 0>, l_aux_angle);
    #local l_newV = vaxis_rotate(l_newV, l_commonaxis, l_angle);
    #local l_newV = l_newV + l_translate;
    l_newV
#end

